/*********************************************************************
 *
 * Condor ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Rajesh Raman.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *********************************************************************/

#include "common.h"
#include "exprTree.h"
#include "sink.h"

#include <iostream.h>

BEGIN_NAMESPACE( classad )

ClassAdUnParser::
ClassAdUnParser()
{
	oldClassAd = false;
}


ClassAdUnParser::
~ClassAdUnParser()
{
}


// should be in same order as OpKind enumeration in common.h
char *ClassAdUnParser::opString[] =
{
    "",     // no-op

    " < ",  // comparisons
    " <= ",
    " != ",
    " == ",
    " >= ",
    " > ",
    " is ",
    " isnt ",

    " +",   // arithmetic
    " -",
    " + ",
    " - ",
    " * ",
    " / ",
    " % ",

    " !",   // logical
    " || ",
    " && ",

    " ~",   // bitwise
    " | ",
    " ^ ",
    " & ",
    " << ",
    " >> ",
    " >>> ",

    " () ", // misc mixfix --- no "single token" representation
    " [] ",
    " ?: "
};

void ClassAdUnParser::
Unparse( std::string &buffer, Value &val )
{
	char	tempBuf[512];
	switch( val.GetType( ) ) {
		case Value::NULL_VALUE: 
			buffer += "(null-value)";
			break;

		case Value::STRING_VALUE: {
			std::string	s;
			val.IsStringValue( s );
			buffer += '"';
			for( std::string::const_iterator itr=s.begin( ); itr!=s.end( ); itr++ ) {
				switch( *itr ) {
					case '\a': buffer += "\\a"; break;
					case '\b': buffer += "\\b"; break;
					case '\f': buffer += "\\f"; break;
					case '\n': buffer += "\\n"; break;
					case '\r': buffer += "\\r"; break;
					case '\t': buffer += "\\t"; break;
					case '\v': buffer += "\\v"; break;
					case '\\': buffer += "\\\\"; break;
					case '\?': buffer += "\\?"; break;
					case '\'': buffer += "\\\'"; break;
					case '\"': buffer += "\\\""; break;

					default:
						if( !isprint( *itr ) ) {
								// print hexadecimal representation
							sprintf( tempBuf, "\\%x", (int)*itr );
							buffer += tempBuf;
						} else {
							buffer += *itr;
						}
				}
			}
			buffer += '"';
			return;
		}
		case Value::INTEGER_VALUE: {
			int	i;
			val.IsIntegerValue( i );
			sprintf( tempBuf, "%d", i );
			buffer += tempBuf;
			return;
		}
		case Value::REAL_VALUE: {
			double	r;
			val.IsRealValue( r );
			sprintf( tempBuf, "%g", r );
			buffer += tempBuf;
			return;
		}
		case Value::BOOLEAN_VALUE: {
			bool b;
			val.IsBooleanValue( b );
//			buffer += b ? "true" : "false";

				// back compatibility only - NAC
			if( oldClassAd ){					// NAC
				buffer += b ? "1" : "0";		// NAC
			}									// NAC
			else{								// NAC
				buffer += b ? "true" : "false";	// NAC
			}									// NAC

			return;
		}
		case Value::UNDEFINED_VALUE: {
			buffer += "undefined";
			return;
		}
		case Value::ERROR_VALUE: {
			buffer += "error";
			return;
		}
		case Value::ABSOLUTE_TIME_VALUE: {
			struct  tm tms;
			char    ascTimeBuf[32], timeZoneBuf[32];
			extern  time_t timezone;
			time_t	asecs;
			val.IsAbsoluteTimeValue( asecs );

				// format:  `Wed Nov 11 13:11:47 1998 (CST) -6:00`
				// we use localtime()/asctime() instead of ctime() 
				// because we need the timezone name from strftime() 
				// which needs a 'normalized' struct tm.  localtime() 
				// does this for us.

				// get the asctime()-like segment; but remove \n
			localtime_r( &asecs, &tms );
			asctime_r( &tms, ascTimeBuf );
			//asctime_r( &tms, ascTimeBuf, 31 );
			ascTimeBuf[24] = '\0';
			buffer += "'";
			buffer += ascTimeBuf;
			buffer += " (";

				// get the timezone name
			if( !strftime( timeZoneBuf, 31, "%Z", &tms ) ) {
				buffer += "<error:strftime>";
				return;
			}
			buffer += timeZoneBuf;
			buffer += ") ";
			
				// output the offSet (use the relative time format)
				// wierd:  POSIX says regions west of GMT have positive
				// offSets.We use negative offSets to not confuse users.
			Value relTime;
			std::string	relTimeBuf;
			relTime.SetRelativeTimeValue( -timezone );
			Unparse( relTimeBuf, relTime );
			buffer += relTimeBuf.substr(1,relTimeBuf.size()-2) + "'";
			return;
		}
		case Value::RELATIVE_TIME_VALUE: {
			time_t	rsecs;
			int		days, hrs, mins, secs;
			val.IsRelativeTimeValue( rsecs );
			buffer += '\'';
			if( rsecs < 0 ) {
				buffer += "-";
				rsecs = -rsecs;
			}
			days = rsecs;
			hrs  = days % 86400;
			mins = hrs  % 3600;
			secs = mins % 60;
			days = days / 86400;
			hrs  = hrs  / 3600;
			mins = mins / 60;

			if( days ) {
				sprintf( tempBuf, "%dd", days );
				buffer += tempBuf;
			}

			sprintf( tempBuf, "%02d:%02d", hrs, mins );
			buffer += tempBuf;
			
			if( secs ) {
				sprintf( tempBuf, ":%02d", secs );
				buffer += tempBuf;
			}
			buffer += '\'';
			return;
		}
		case Value::CLASSAD_VALUE: {
			ClassAd *ad;
			std::vector< std::pair<std::string,ExprTree*> > attrs;
			val.IsClassAdValue( ad );
			ad->GetComponents( attrs );
			UnparseAux( buffer, attrs );
			return;
		}
		case Value::LIST_VALUE: {
			const ExprList *el;
			std::vector<ExprTree*> exprs;
			val.IsListValue( el );
			el->GetComponents( exprs );
			UnparseAux( buffer, exprs );
			return;
		}
	}
}


void ClassAdUnParser::
Unparse( std::string &buffer, ExprTree *tree )
{
	if( !tree ) {
		buffer = "<error:null expr>";
		return;
	}

	switch( tree->GetKind( ) ) {
		case ExprTree::LITERAL_NODE: {
			Value				val;
			Value::NumberFactor	factor;
			((Literal*)tree)->GetComponents( val, factor );
			UnparseAux( buffer, val, factor );
			return;
		}

		case ExprTree::ATTRREF_NODE: {
			ExprTree *expr;
			std::string	ref;
			bool	absolute;
			((AttributeReference*)tree)->GetComponents( expr, ref, absolute );
			UnparseAux( buffer, expr, ref, absolute );
			return;
		}

		case ExprTree::OP_NODE: {
			Operation::OpKind	op;
			ExprTree			*t1, *t2, *t3;
			((Operation*)tree)->GetComponents( op, t1, t2, t3 );
			UnparseAux( buffer, op, t1, t2, t3 );
			return;
		}

		case ExprTree::FN_CALL_NODE: {
			std::string fnName;
			std::vector<ExprTree*> args;
			((FunctionCall*)tree)->GetComponents( fnName, args );
			UnparseAux( buffer, fnName, args );
			return;
		}

		case ExprTree::CLASSAD_NODE: {
			std::vector< std::pair<std::string, ExprTree*> > attrs;
			((ClassAd*)tree)->GetComponents( attrs );
			UnparseAux( buffer, attrs );
			return;
		}

		case ExprTree::EXPR_LIST_NODE: {
			cout << "warning: EXPR_LIST_NODE" << endl;
			std::vector<ExprTree*> exprs;
			((ExprList*)tree)->GetComponents( exprs );
			UnparseAux( buffer, exprs );
			return;
		}

		default:
			buffer = "";
			CondorErrno = ERR_BAD_EXPRESSION;
			CondorErrMsg = "unknown expression type";
			return;
	}
}
				


void ClassAdUnParser::
UnparseAux( std::string &buffer, Value &val, Value::NumberFactor factor )
{
	Unparse( buffer, val );
	if( val.IsNumber( ) && factor != Value::NO_FACTOR ) {
		buffer += (factor==Value::B_FACTOR)?"B"  :
					(factor==Value::K_FACTOR)?"K": 
					(factor==Value::M_FACTOR)?"M":
					(factor==Value::G_FACTOR)?"G": 
					(factor==Value::T_FACTOR)?"T": 
					"<error:bad factor>";
	}
	return;
}

void ClassAdUnParser::
UnparseAux( std::string &buffer, ExprTree *expr, std::string &attrName, bool absolute )
{
	if( expr ) {
		Unparse( buffer, expr );
		buffer += "." + attrName;
		return;
	}
	if( absolute ) buffer += ".";
	buffer += attrName;
}

void ClassAdUnParser::
UnparseAux(std::string &buffer, Operation::OpKind op, ExprTree *t1, ExprTree *t2, 
	ExprTree *t3)
{
		// case 0: parentheses op
	if( op==Operation::PARENTHESES_OP ) {
		buffer += "( ";
		Unparse( buffer, t1 );
		buffer += " )";
		return;
	}
		// case 1: check for unary ops
	if( op==Operation::UNARY_PLUS_OP || op==Operation::UNARY_MINUS_OP || 
			op==Operation::LOGICAL_NOT_OP || op==Operation::BITWISE_NOT_OP ) {
		buffer += opString[op];
		Unparse( buffer, t1 );
		return;
	}
		// case 2: check for ternary op
	if( op==Operation::TERNARY_OP ) {
		Unparse( buffer, t1 );
		buffer += " ? ";
		Unparse( buffer, t2 );
		buffer += " : ";
		Unparse( buffer, t3 );
		return;
	}
		// case 3: check for subscript op
	if( op==Operation::SUBSCRIPT_OP ) {
		Unparse( buffer, t1 );
		buffer += '[';
		Unparse( buffer, t2 );
		buffer += ']';
		return;
	}

		// back compatibility only - NAC
	if( oldClassAd ) {
		Unparse( buffer, t1 );
		if( op==Operation::META_EQUAL_OP ) {
			buffer += " =?= ";
		}
		else if( op==Operation::META_NOT_EQUAL_OP ) {
			buffer += " =!= ";
		}
		else {
			buffer += opString[op];
		}
		Unparse( buffer, t2 );
		return;
	}

		// all others are binary ops
	Unparse( buffer, t1 );
	buffer += opString[op];
	Unparse( buffer, t2 );
}

void ClassAdUnParser::
UnparseAux( std::string &buffer, std::string &fnName, std::vector<ExprTree*>& args )
{
	std::vector<ExprTree*>::const_iterator	itr;

	buffer += fnName + "(";
	for( itr=args.begin( ); itr!=args.end( ); itr++ ) {
		Unparse( buffer, *itr );
		if( itr+1 != args.end( ) ) buffer += ',';
	}
	buffer += ")";
}

void ClassAdUnParser::
UnparseAux( std::string &buffer, std::vector< std::pair<std::string,ExprTree*> >& attrs )
{
	std::vector< std::pair<std::string,ExprTree*> >::const_iterator itr;

	buffer += "[ ";
	for( itr=attrs.begin( ); itr!=attrs.end( ); itr++ ) {
		buffer += itr->first + " = ";
		Unparse( buffer, itr->second );
		if( itr+1 != attrs.end( ) ) buffer += "; ";
	}
	buffer += " ]";
}


void ClassAdUnParser::
UnparseAux( std::string &buffer, std::vector<ExprTree*>& exprs )
{
	std::vector<ExprTree*>::const_iterator	itr;

	buffer += "{ ";
	for( itr=exprs.begin( ); itr!=exprs.end( ); itr++ ) {
		Unparse( buffer, *itr );
		if( itr+1 != exprs.end( ) ) buffer += ',';
	}
	buffer += " }";
}

// back compatibility only - NAC
void ClassAdUnParser::
SetOldClassAd( bool b )
{
	oldClassAd = b;
}

bool ClassAdUnParser::
GetOldClassAd()
{
	return oldClassAd;
}


// PrettyPrint object implementation
PrettyPrint::
PrettyPrint( )
{
	classadIndent = 4;
	listIndent = 3;
	wantStringQuotes = true;
	minimalParens = false;
	indentLevel = 0;
}


PrettyPrint::
~PrettyPrint( )
{
}

void PrettyPrint::
SetClassAdIndentation( int len )
{
	classadIndent = len;
}

int PrettyPrint::
GetClassAdIndentation( )
{
	return( classadIndent );
}

void PrettyPrint::
SetListIndentation( int len )
{
	listIndent = len;
}

int PrettyPrint::
GetListIndentation( )
{
	return( listIndent );
}

void PrettyPrint::
SetWantStringQuotes( bool b )
{
	wantStringQuotes = b;
}

bool PrettyPrint::
GetWantStringQuotes( )
{
	return( wantStringQuotes );
}

void PrettyPrint::
SetMinimalParentheses( bool b )
{
	minimalParens = b;
}

bool PrettyPrint::
GetMinimalParentheses( )
{
	return( minimalParens );
}

void PrettyPrint::
UnparseAux(std::string &buffer,Operation::OpKind op,ExprTree *op1,ExprTree *op2,
	ExprTree *op3)
{
	if( !minimalParens ) {
		return( ClassAdUnParser::UnparseAux( buffer, op, op1, op2, op3 ) );
	}

		// case 0: parentheses op
	if( op==Operation::PARENTHESES_OP ) {
		Unparse( buffer, op1 );
		return;
	}
		// case 1: check for unary ops
	if( op==Operation::UNARY_PLUS_OP || op==Operation::UNARY_MINUS_OP || 
			op==Operation::LOGICAL_NOT_OP || op==Operation::BITWISE_NOT_OP ) {
		buffer += opString[op];
		Unparse( buffer, op1 );
		return;
	}
		// case 2: check for ternary op
	if( op==Operation::TERNARY_OP ) {
		Unparse( buffer, op1 );
		buffer += " ? ";
		Unparse( buffer, op2 );
		buffer += " : ";
		Unparse( buffer, op3 );
		return;
	}
		// case 3: check for subscript op
	if( op==Operation::SUBSCRIPT_OP ) {
		Unparse( buffer, op1 );
		buffer += '[';
		Unparse( buffer, op2 );
		buffer += ']';
		return;
	}
		// all others are binary ops
	Operation::OpKind	top;
	ExprTree			*t1, *t2, *t3;

	if( op1->GetKind( ) == ExprTree::OP_NODE ) {
		((Operation*)op1)->GetComponents( top, t1, t2, t3 );
		if( Operation::PrecedenceLevel(top)<Operation::PrecedenceLevel(op) ) {
			buffer += " ( ";
			UnparseAux( buffer, top, t1, t2, t3 );
			buffer += " ) ";
		}
	} else {
		Unparse( buffer, t1 );
	}
	buffer += opString[op];
	if( op2->GetKind( ) == ExprTree::OP_NODE ) {
		((Operation*)op2)->GetComponents( top, t1, t2, t3 );
		if( Operation::PrecedenceLevel(top)<Operation::PrecedenceLevel(op) ) {
			buffer += " ( ";
			UnparseAux( buffer, top, t1, t2, t3 );
			buffer += " ) ";
		}
	} else {
		Unparse( buffer, t1 );
	}
}


void PrettyPrint::
UnparseAux( std::string &buffer, std::vector< std::pair<std::string,ExprTree*> >& attrs )
{
	std::vector< std::pair<std::string, ExprTree*> >::iterator	itr;

	if( classadIndent > 0 ) {
		indentLevel += classadIndent;
		buffer += '\n' + std::string( indentLevel, ' ' ) + '[';
		indentLevel += classadIndent;
	} else {
		buffer += "[ ";
	}
	for( itr=attrs.begin( ); itr!=attrs.end( ); itr++ ) {
		if( classadIndent > 0 ) {
			buffer += '\n' + std::string( indentLevel, ' ' );
		}
		buffer += itr->first + " = ";
		Unparse( buffer, itr->second );
		if( itr+1 != attrs.end( ) ) buffer += "; ";
	}
	if( classadIndent > 0 ) {
		indentLevel -= classadIndent;
		buffer += '\n' + std::string( indentLevel, ' ' ) + ']';
		indentLevel -= classadIndent;
	} else {
		buffer += " ]";
	}
}


void PrettyPrint::
UnparseAux( std::string &buffer, std::vector<ExprTree*>& exprs )
{
	std::vector<ExprTree*>::iterator	itr;

	if( listIndent > 0 ) {
		indentLevel += listIndent;
		buffer += '\n' + std::string( indentLevel, ' ' ) + '{';
		indentLevel += listIndent;
	} else {
		buffer += "{ ";
	}
	for( itr=exprs.begin( ); itr!=exprs.end( ); itr++ ) {
		if( listIndent > 0 ) {
			buffer += '\n' + std::string( indentLevel, ' ' );
		}
		ClassAdUnParser::Unparse( buffer, *itr );
		if( itr+1 != exprs.end( ) ) buffer += ',';
	}
	if( listIndent > 0 ) {
		indentLevel -= listIndent;
		buffer += '\n' + std::string( indentLevel, ' ' ) + '}';
		indentLevel -= listIndent;
	} else {
		buffer += " }";
	}
}


// sexprExtract object implementation
sexprExtract::sexprExtract( )
{
  desired_equation_type = NOT_SET;
  SetOpLevel = 0;
}


sexprExtract::~sexprExtract( )
{
}

void
sexprExtract::SetDesiredEquationType(Equation_Type t)
{
   desired_equation_type = t;
   return;
}

// should be in same order as OpKind enumeration in common.h
char*
sexprExtract::opString[] =
{
    "",     // no-op

    " < ",  // comparisons
    " <= ",
    " nequal? ",
    " equal? ",
    " >= ",
    " > ",
    " equal? ",       // -- Problem for interpreter, treat "is" as "equal"
    " nequal? ",      // -- Problem for interpreter, treat "isnt" as "nequal"

    " +",   // arithmetic
    " -",
    " + ",		// why these have trailing " ", but first 2 don't?
    " - ",
    " * ",
    " / ",
    " remainder ",

    " not ",   // logical
    " or ",
    " and ",

    " ~",   // bitwise   -- Problems for interpreter??
    " | ",
    " ^ ",
    " & ",
    " << ",
    " >> ",
    " >>> ",

    " () ", 		 // -- should never see this (?)
    " [] ",		 // -- Problem for interpreter
    " if "		 // -- ( a ? b : c ) ->  ( if a b c )
};

void 
sexprExtract::Unparse( std::string &buffer, 
                       Value &val )
{
	char	tempBuf[512];
	switch( val.GetType( ) ) {
	  case Value::NULL_VALUE: 
		buffer += "(null-value)";
		break;

	  case Value::STRING_VALUE: {
		std::string	s;
		val.IsStringValue( s );
		buffer += '"';
		for( std::string::const_iterator itr=s.begin( ); 
				                 itr!=s.end( ); itr++ ) {
		  switch( *itr ) {
			case '\a': buffer += "\\a";   break;
			case '\b': buffer += "\\b";   break;
			case '\f': buffer += "\\f";   break;
			case '\n': buffer += "\\n";   break;
			case '\r': buffer += "\\r";   break;
			case '\t': buffer += "\\t";   break;
			case '\v': buffer += "\\v";   break;
			case '\\': buffer += "\\\\";  break;
			case '\?': buffer += "\\?";   break;
			case '\'': buffer += "\\\'";  break;
			case '\"': buffer += "\\\"";  break;

			default:
				if( !isprint( *itr ) ) {
				  sprintf( tempBuf, "\\%x", (int)*itr );
				  buffer += tempBuf;
				} else {
				  buffer += *itr;
				}
			}
		  }
		buffer += '"';
		return;
	    }
	  case Value::INTEGER_VALUE: {
		int	i;
		val.IsIntegerValue( i );
		sprintf( tempBuf, "%d", i );
		buffer += tempBuf;
		return;
	    }
	  case Value::REAL_VALUE: {
		double	r;
		val.IsRealValue( r );
		sprintf( tempBuf, "%g", r );
		buffer += tempBuf;
		return;
	    }
	  case Value::BOOLEAN_VALUE: {
		bool b;
		val.IsBooleanValue( b );
		buffer += b ? "true" : "false";
		return;
	    }
 	  case Value::UNDEFINED_VALUE: {
		buffer += "undefined";
		return;
	    }
	  case Value::ERROR_VALUE: {
		buffer += "error";
		return;
	    }
	  case Value::ABSOLUTE_TIME_VALUE: {
		struct  tm tms;
		char    ascTimeBuf[32], timeZoneBuf[32];
		extern  time_t timezone;
		time_t	asecs;
		val.IsAbsoluteTimeValue( asecs );
		localtime_r( &asecs, &tms );
		asctime_r( &tms, ascTimeBuf );
		ascTimeBuf[24] = '\0';
		buffer += "'";
		buffer += ascTimeBuf;
		buffer += " (";

		if( !strftime( timeZoneBuf, 31, "%Z", &tms ) ) {
			buffer += "<error:strftime>";
			return;
		}
		buffer += timeZoneBuf;
		buffer += ") ";
			
		Value relTime;
		std::string	relTimeBuf;
		relTime.SetRelativeTimeValue( -timezone );
		Unparse( relTimeBuf, relTime );
		buffer += relTimeBuf.substr(1,relTimeBuf.size()-2) + "'";
		return;
	    }
	  case Value::RELATIVE_TIME_VALUE: {
		time_t	rsecs;
		int		days, hrs, mins, secs;
		val.IsRelativeTimeValue( rsecs );
		buffer += '\'';
		if( rsecs < 0 ) {
			buffer += "-";
			rsecs = -rsecs;
		}
		days = rsecs;
		hrs  = days % 86400;
		mins = hrs  % 3600;
		secs = mins % 60;
		days = days / 86400;
		hrs  = hrs  / 3600;
		mins = mins / 60;

		if( days ) {
			sprintf( tempBuf, "%dd", days );
			buffer += tempBuf;
		}

		sprintf( tempBuf, "%02d:%02d", hrs, mins );
		buffer += tempBuf;
			
		if( secs ) {
			sprintf( tempBuf, ":%02d", secs );
			buffer += tempBuf;
		}
		buffer += '\'';
		return;
	    }
	  case Value::CLASSAD_VALUE: {
		ClassAd *ad;
		std::vector< std::pair<std::string,ExprTree*> > attrs;
		val.IsClassAdValue( ad );
		ad->GetComponents( attrs );
		UnparseAux( buffer, attrs );
		return;
	    }
	  case Value::LIST_VALUE: {
		const ExprList *el;
		std::vector<ExprTree*> exprs;
		val.IsListValue( el );
		el->GetComponents( exprs );
		UnparseAux( buffer, exprs );
		return;
	    }
	}
}


void 
sexprExtract::Unparse( std::string &buffer, ExprTree *tree )
{
	if( !tree ) {
		buffer = "<error:null expr>";
		return;
	}

	switch( tree->GetKind( ) ) {
		case ExprTree::LITERAL_NODE: {
			Value				val;
			Value::NumberFactor	factor;
			((Literal*)tree)->GetComponents( val, factor );
			UnparseAux( buffer, val, factor );
			return;
		}

		case ExprTree::ATTRREF_NODE: {
			ExprTree *expr;
			std::string	ref;
			bool	absolute;
			((AttributeReference*)tree)->GetComponents( expr, 
								    ref, 
								    absolute );
			UnparseAux( buffer, expr, ref, absolute );
			return;
		}

		case ExprTree::OP_NODE: {
			Operation::OpKind	op;
			ExprTree			*t1, *t2, *t3;
			((Operation*)tree)->GetComponents( op, t1, t2, t3 );
			UnparseAux( buffer, op, t1, t2, t3 );
			return;
		}

		case ExprTree::FN_CALL_NODE: {
			std::string fnName;
			std::vector<ExprTree*> args;
			((FunctionCall*)tree)->GetComponents( fnName, args );
			if ( args.size() != 1 ) {
			  buffer.erase();
			  buffer = 
		 	       "ERROR: multi-arg function calls unsupported";
			  return;
			}

			if ( fnName == "Sum" ) {
			  fnName = "(class_ads_setsum (quote ";
			  SetOpLevel++;
			  UnparseAux( buffer, fnName, args );
			  SetOpLevel--;
			  return;
			} else if ( fnName == "Min" ) {
			  fnName = "(class_ads_setmin (quote ";
			  SetOpLevel++;
			  UnparseAux( buffer, fnName, args );
			  SetOpLevel--;
			  return;
			} else if ( fnName == "Max" ) {
			  fnName = "(class_ads_setmax (quote ";
			  SetOpLevel++;
			  UnparseAux( buffer, fnName, args );
			  SetOpLevel--;
			  return;
			} else if ( fnName == "SizeOfSet" ) {
			  fnName = "(grads_number_of_resources)";
			  SetOpLevel++;
			  buffer += fnName;
			  SetOpLevel--;
			  return;
			} else {
			  buffer.erase();
			  buffer = "ERROR: general function calls unsupported";
			  return;
			}
		}

		case ExprTree::CLASSAD_NODE: {
			std::vector< std::pair<std::string, ExprTree*> > attrs;
			((ClassAd*)tree)->GetComponents( attrs );
			UnparseAux( buffer, attrs );
			return;
		}

		case ExprTree::EXPR_LIST_NODE: {
			cout << "warning: EXPR_LIST_NODE" << endl;
			std::vector<ExprTree*> exprs;
			((ExprList*)tree)->GetComponents( exprs );
			UnparseAux( buffer, exprs );
			return;
		}

		default:
			buffer.erase();
			buffer = "ERROR: unknown expression type";
			return;
	}
}
				

void 
sexprExtract::UnparseAux( std::string &buffer, 
			  Value &val, 
			  Value::NumberFactor factor )
{
	Unparse( buffer, val );
	if( val.IsNumber( ) && factor != Value::NO_FACTOR ) {
		buffer += (factor==Value::B_FACTOR)?"B"  :
					(factor==Value::K_FACTOR)?"K": 
					(factor==Value::M_FACTOR)?"M":
					(factor==Value::G_FACTOR)?"G": 
					(factor==Value::T_FACTOR)?"T": 
					"<error:bad factor>";
	}
	return;
}


void 
sexprExtract::UnparseAux( std::string &buffer, 
                          ExprTree *expr, 
                          std::string &attrName, 
                          bool absolute )
{
	if( expr ) {
		Unparse( buffer, expr );
		buffer += "." + attrName;
		return;
	}
	if( absolute ) buffer += ".";
        // we now do an ugly search for this attrName in our 
        // known equations
        // we want to compare to the string and, if equal, then
        // replace it with the s-expr for the ExprTree*
	std::vector< std::pair<std::string,ExprTree*> >::const_iterator itr;
	bool found = false;
        std::string side_buffer;
	for ( itr=equations.begin( ); itr!=equations.end( ); itr++ ) {
		if ( itr->first == attrName ) {
			found = true;
			Unparse( side_buffer, itr->second );
			buffer += side_buffer;
			return;
		}
        }
	if ( !found) {
	  if ( (attrName.size() > 6)              &&
	       (attrName.substr(0,6) == "other_") ) {
            if ( SetOpLevel < 1 ) {
	      buffer.erase();
	      buffer = "ERROR: set_variable outside of set_operation!";
	      return;
            }
	    buffer += 
                 "(grads_get_resource_property grads_resource_index (quote ";
	    buffer += attrName;
	    buffer += "))";
	  } else {
            buffer += attrName;
	  }
	}
	return;
}

void 
sexprExtract::UnparseAux( std::string &buffer,
                          Operation::OpKind op,
                          ExprTree *op1, ExprTree *op2, ExprTree *op3)
{
                // case 0: parentheses op
        if( op==Operation::PARENTHESES_OP ) {
		// ignore these Parens
                Unparse( buffer, op1 );
                return;
        }
                // case 1: check for unary ops
        if( op==Operation::UNARY_PLUS_OP || op==Operation::UNARY_MINUS_OP ||
            op==Operation::LOGICAL_NOT_OP || op==Operation::BITWISE_NOT_OP ) {
                buffer += opString[op];
                Unparse( buffer, op1 );
                return;
        }
                // case 2: check for ternary op     
        if( op==Operation::TERNARY_OP ) {
        	buffer += "(";
        	buffer += opString[op];
        	Unparse( buffer, op1 );
        	buffer += " ";
        	Unparse( buffer, op2 );
        	buffer += " ";
        	Unparse( buffer, op3 );
        	buffer += " )";
                return;
        }
                // case 3: check for subscript op  
        if( op==Operation::SUBSCRIPT_OP ) {
		cout << "Unsupported SUBSCRIPT_OP found" << endl;
                return;
        }

                // all others are binary ops
        buffer += "(";
        buffer += opString[op];
        Unparse( buffer, op1 );
        buffer += " ";
        Unparse( buffer, op2 );
        buffer += " )";
}


void 
sexprExtract::UnparseAux( std::string &buffer, 
                          std::vector< 
                               std::pair<std::string,ExprTree*> >& attrs )
{
	std::vector< std::pair<std::string, ExprTree*> >::iterator	itr;

        // if I wasn't trying to use the classad ExprTree, it wouldn't
        // be necessary to build lists of what identifiers are equations
        // and what are variables - I would have created a different tree
        // structure.  oh well...
        for( itr=attrs.begin( ); itr!=attrs.end( ); itr++ ) {
          if ( itr->second->GetKind() != ExprTree::LITERAL_NODE ) {
                equations.push_back(*itr);
          } else {
                variables.push_back(*itr);
          }
        }

	let_block = " ( let (";
	for( itr=attrs.begin( ); itr!=attrs.end( ); itr++ ) {
          if ( itr->second->GetKind() == ExprTree::LITERAL_NODE ) {
		let_block += " (" + itr->first + " ";
		Unparse( let_block, itr->second );
		let_block += ")";
          }
        }
	if ( let_block != " ( let (" ) {
          let_block += " ) ";
	} else {
	  let_block = "";
	}
	buffer += let_block;
	for( itr=attrs.begin( ); itr!=attrs.end( ); itr++ ) {
          if ( itr->second->GetKind() != ExprTree::LITERAL_NODE ) {
	    if ( (itr->first == "rank") &&
	         (desired_equation_type == RANK) ) 
		         Unparse( buffer, itr->second );
	    if ( (itr->first == "requirements") &&
	         (desired_equation_type == REQMTS) )
			 Unparse( buffer, itr->second );
          }
	}
        if ( let_block != "" ) buffer += " )";
}


void 
sexprExtract::UnparseAux( std::string &buffer, 
                          std::vector<ExprTree*>& exprs )
{
	std::vector<ExprTree*>::iterator	itr;

	buffer += "{ ";
	for( itr=exprs.begin( ); itr!=exprs.end( ); itr++ ) {
		ClassAdUnParser::Unparse( buffer, *itr );
		if( itr+1 != exprs.end( ) ) buffer += ',';
	}
	buffer += " }";
}

void 
sexprExtract::UnparseAux( std::string &buffer, 
                          std::string &fnName, 
                          std::vector<ExprTree*>& args )
{
	  buffer += fnName;
	  buffer += let_block;
	  Unparse(buffer, args[0]); 
          buffer += "))";
	  if ( let_block.size()  != 0 ) buffer += ")";
}


END_NAMESPACE // classad
