/* $Id: tCA_Mapper.cc,v 1.5 2002/10/10 13:53:20 mazina Exp $
 */


/* ========================================================================= */
/*                                                                           */
/*    Copyright 2003,                                                        */
/*      Rice University,                                                     */
/*      University of California-San Diego,                                  */
/*      University of California-Santa Barbara,                              */
/*      University of Chicago,                                               */
/*      University of Houston,                                               */
/*      University of Illinois,                                              */
/*      University of Southern California-Information Sciences Institute,    */
/*      University of Tennessee-Knoxville.                                   */
/*                                                                           */
/*    All Rights Reserved.                                                   */
/*                                                                           */
/* See GrADSoft_Authors file in the GrADSoft top-level directory             */
/*                           for a list of developers.                       */
/*                                                                           */ 
/* Permission to use, copy, modify and distribute any part of this source    */
/* file of the Grid Application Development Software (GrADS) project for     */
/* educational, research and non-profit purposes, without fee, and without a */
/* written agreement is hereby granted, provided that the above copyright    */
/* notice, this paragraph and the following two paragraphs appear in all     */
/* copies.                                                                   */
/*                                                                           */
/*                                                                           */
/* IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS, EITHER JOINTLY OR    */
/* SEPARATELY, BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,         */
/* INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT */
/* OF THE USE OF THIS SOFTWARE, EVEN IF THEY HAVE BEEN ADVISED OF THE        */
/* POSSIBILITY OF SUCH DAMAGE.                                               */
/*                                                                           */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED HEREIN ON AN "AS IS" BASIS, AND THE AUTHORS AND */
/* COPYRIGHT HOLDERS HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,     */
/* UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE AUTHORS AND COPYRIGHT        */
/* HOLDERS MAKE NO REPRESENTATIONS AND EXTEND NO WARRANTIES OF ANY KIND,     */
/* EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     */
/* WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR     */
/* THAT THE USE OF THIS SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR  */
/* OTHER RIGHTS.                                                             */
/*                                                                           */
/* ========================================================================= */

#include <stdio.h>

#include "syslog.h"
#include "tCA_Mapper.h"
#include "CA2GrADSiface.h"
#include "helpers.h"      // For sort function

///////////////////////////////////////////////////////////////////////
// translated classAd Mapper methods

// std constructor
catMapper::catMapper(RSSM* inputRSSM, AttListPtr ConfigListPtr) :
                            Mapper(inputRSSM, ConfigListPtr) {

  Syslog::setVerbosity(WARNING);

  Syslog::log(DEBUG, "catMapper", "constructor called");

  copRSSM = inputRSSM;
  ProbConfigListPtr = ConfigListPtr;

}

catMapper::~catMapper() {

  Syslog::log(DEBUG, "catMapper", "destructor called");

}

int catMapper::GetMapping( int numChosen,
                           int* indices,
                           RSSM* rssm,
                           RSE* rse,
                           ResourceBag* gridInfo,
                           Network* network,
                           Schedule* sched ) {
  int i, memValid, compValid, cpusValid;
  int *sortIndex;
  int *tempIndices;
  double *tempSiteID;    // a double for the use of sort function?

  Syslog::log(DEBUG, "catMapper", "GetMapping called.");

  // ORDERING OF PROCESSORS

  // This mapper orders the processors based on nearness to get a
  // communication-efficient processor arrangement.
  tempSiteID = new double[numChosen];
  sortIndex = new int[numChosen];
  tempIndices = new int[numChosen];

  for(i=0; i < numChosen; i++){
    tempIndices[i] = indices[i];
    tempSiteID[i] = gridInfo->siteID[ indices[i] ];
  }

  GetIndex( tempSiteID, sortIndex, numChosen, ASCENDING );

  // Begin filling in schedule structure
  sched->size = numChosen;

  for( i = 0; i < numChosen; i++ ) { 
    sched->indices[i] = tempIndices[ sortIndex[i] ];
  }

  // WORK ALLOCATION
  // This mapper does even work allocation

  for( i = 0; i < numChosen; i++ ) {
    sched->workAlloc[i] = 1.0 / (double) numChosen; 
  }

  Syslog::log(DEBUG, "catMapper", "work allocation done.");

  // set-up for evaluation of requirements and rank

  EnvContextPtr dvp = new ResourceDataVector();
  if ( ! dvp->LoadData(sched, gridInfo, network) ) {
    Syslog::log(ERROR, "catMapper", "ResourceDataVector.LoadData() fails.");
    return(-1);
  }
  Syslog::log(DEBUG, "catMapper", "finished resource LoadData().");

  catRSSM *rssmPtr = (catRSSM *) rssm;
  bool evalSuccess;
  bool CA_requirements_met;
  double CA_rank_value;

  rssmPtr->Reqmts_Function->setEnvContextPtr(dvp);
  rssmPtr->Reqmts_Function->setEnvFuncPtr(&GetResourceProperty);
  CA_requirements_met = 
	          rssmPtr->Reqmts_Function->evalBoolFunction(&evalSuccess);
  if ( ! evalSuccess ) {
    Syslog::log(ERROR, "catMapper", "CA Reqmts_Function fails");
    return(-1);
  }
  if ( ! CA_requirements_met ) {
    Syslog::log(INFO, "catMapper", "CA Reqmts not met");
    return(-1);
  }

  Syslog::log(DEBUG, "catMapper", "finished eval Reqmts.");

#define OPTION1_FOR_EXEC_COST

#ifdef OPTION1_FOR_EXEC_COST
  // most efficient, but cuts RSE object out totally by directly calling
  // the Rank functionalAttribute.
  //
  rssmPtr->Rank_Function->setEnvContextPtr(dvp);
  rssmPtr->Rank_Function->setEnvFuncPtr(&GetResourceProperty);
  CA_rank_value = rssmPtr->Rank_Function->evalFloatFunction(&evalSuccess);
  if ( ! evalSuccess ) {
    Syslog::log(ERROR, "catMapper", "CA Rank_Function (Opt1) fails");
    return(-1);
  }
#endif
#ifdef OPTION2_FOR_EXEC_COST
  // uses RSE object, but also only constructs one ResourceDataVector
  // (here, in the mapper) which is passed to the RSE.
  //
  catRSE *rsePtr = (catRSE *) rse;
  if ( rsePtr->GetExecutionTimeCost(dvp, &CA_rank_value) == 1 ) {
    Syslog::log(ERROR, "catMapper", "CA Rank_Function (Opt2) fails");
    return(-1);
  }
#endif
#ifdef OPTION3_FOR_EXEC_COST
  // most general use of RSE following original interface, but most 
  // inefficient as both mapper and RSE construct 
  // identical ResourceDataVector objects.
  //
  catRSE *rsePtr = (catRSE *) rse;
  if ( rsePtr->GetExecutionTimeCost(sched, gridInfo, 
			            network, &CA_rank_value) == 1 ) {
    Syslog::log(ERROR, "catMapper", "CA Rank_Function (Opt3) fails");
    return(-1);
  }
#endif

  Syslog::log(DEBUG, "catMapper", "finished eval Rank.");

  sched->execCost = CA_rank_value;

  // CALCULATE AGGREGATE CHARACTERISTICS OF SCHEDULE 
  // Scheduler can use if desired to compare scheds
  sched->aggComp = 0;
  sched->aggMem = 0;
  sched->aggCpus = 0;

  compValid = 1; memValid = 1; cpusValid = 1;

  for( i = 0; i < numChosen; i++ ) {
    if( ((gridInfo->comp[sched->indices[i]]) > 0) &&
        compValid ) {
      // should calculate the sum
      sched->aggComp += gridInfo->comp[sched->indices[i]];
    } else {
      compValid = 0;
    }
    if( ((gridInfo->mem[sched->indices[i]]) > 0 ) && 
        memValid ) {
      sched->aggMem += gridInfo->mem[sched->indices[i]];
    } else {
      memValid = 0;
    }
    if( ((gridInfo->cpus[sched->indices[i]]) > 0 ) && 
        cpusValid ) {
      sched->aggCpus += gridInfo->cpus[sched->indices[i]];
    } else {
      cpusValid = 0;
    }
  }

  if( !compValid ) sched->aggComp = -1.0;
  if( !memValid  ) sched->aggMem = -1.0;
  if( !cpusValid ) sched->aggCpus = -1.0;
  
  return 0;
}


Mapper *CreateMapperObject(RSSM* inputRSSM, AttListPtr ConfigListPtr)
{
  catMapper *catMapperPtr = new catMapper(inputRSSM, ConfigListPtr);
  return((Mapper *)catMapperPtr);
}

